/* asm41 - HP-41 assembler */

/* Copyright (C) 2015 Kari Pasanen */
/* Full copyright notice in document file asm41.html */
/* Rev. 1.2 published 2015-02-16 */

/* Usage: asm41 asmfile */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "asm41.h"

#define MAX_LINE_LEN 79
#define COMMENT_COLUMN 34
#define MAX_SYMBOL_LENGTH 10
#define NUM_SYMBOLS 2000
#define NUM_SYMBOL_USES 4000

static unsigned int org;
static unsigned int code;
static char second_pass;
static char quiet;
static char org_defined;
static char directive;
static char added;
static char *instrP;
static char *paramP;
static char *commentP;
static char *addedinstrP;
static char *addedparamP;
static char *addedcommentP;
static char name[MAX_SYMBOL_LENGTH + 1];
static char line[MAX_LINE_LEN + 1];
static char output[MAX_LINE_LEN + COMMENT_COLUMN + 1];
static char addinstr[8];
static char addparam[8];
static char addcomment[12];
static FILE *outputP;

static int pagenum;
static unsigned int ctrlbits;
static unsigned char object[0x2000];

static struct symbol {
        struct symbol_use *useP;
        char name[MAX_SYMBOL_LENGTH + 1];
        char global;
        char satisfied;
        unsigned int value;
} symbol_table[NUM_SYMBOLS];

static struct symbol_use {
        struct symbol_use *next_useP;
        struct symbol *symbolP;
        unsigned int location;
} symbol_use_table[NUM_SYMBOL_USES];

static unsigned int first_free_symbol;
static unsigned int first_free_symbol_use;

static void handle_file(char *);


/* Error exit functions */

static void file_open_read_error(
        char *filenameP) {

        fprintf(stderr, "Could not open %s for read\n", filenameP);
        exit(100);
}

static void too_long_line(void) {

        fprintf(stderr, "Too long line\n");
        exit(101);
}

static void brace_error(void) {

        fprintf(stderr, "Brace error\n");
        exit(102);
}

static void extra_second_parameter(void) {

        fprintf(stderr, "Extra parameter in %s %s\n", instrP, paramP);
        exit(103);
}

static void too_long_hex_number(void) {

        fprintf(stderr, "Too long hexadecimal number\n");
        exit(104);
}

static void illegal_character(void) {

        fprintf(stderr, "Illegal character\n");
        exit(105);
}

static void empty_directive(void) {

        fprintf(stderr, "Empty directive\n");
        exit(106);
}

static void unknown_directive(void) {

        fprintf(stderr, "Unknown directive %s\n", instrP);
        exit(107);
}

static void before_ORG(void) {

        if (directive)
                fprintf(stderr, "%%");

        fprintf(stderr, "%s before %%ORG\n", instrP);
        exit(108);
}

static void instruction_not_found(void) {

        fprintf(stderr, "Instruction %s not found\n", instrP);
        exit(109);
}

static void missing_parameter(void) {

        fprintf(stderr, "Missing parameter in %s\n", instrP);
        exit(110);
}

static void extra_parameter(void) {

        fprintf(stderr, "Extra parameter in %s\n", instrP);
        exit(111);
}

static void org_rel_syntax_error(void) {

        fprintf(stderr, "Syntax error in org-relative %s\n", instrP);
        exit(112);
}

static void too_long_jump(void) {

        fprintf(stderr, "Too long relative jump\n");
        exit(113);
}

static void parameter_not_found(void) {

        fprintf(stderr, "Parameter %s not found\n", paramP);
        exit(114);
}

static void too_long_symbol(
        char *nameP) {

        fprintf(stderr, "Too long symbol %s\n", nameP);
        exit(115);
}

static void duplicate_symbol(
        char *nameP) {

        fprintf(stderr, "Duplicate symbol %s\n", nameP);
        exit(116);
}

static void symbol_table_full(void) {

        fprintf(stderr, "Symbol table full\n");
        exit(117);
}

static void symbol_use_table_full(void) {

        fprintf(stderr, "Symbol use table full\n");
        exit(118);
}

static void unsatisfied_symbol(
        char *nameP) {

        fprintf(stderr, "Unsatisfied symbol %s\n", nameP);
        exit(119);
}

static void file_open_write_error(
        char *filenameP) {

        fprintf(stderr, "Could not open %s for write\n", filenameP);
        exit(120);
}

static void too_long_page_number(void) {

        fprintf(stderr, "Too long page number\n");
        exit(121);
}

static void object_page_not_defined(void) {

        fprintf(stderr, "Object page not defined\n");
        exit(122);
}

static void overwriting_object_data(void) {

        fprintf(stderr, "Overwriting object data\n");
        exit(123);
}


/* Symbol table and symbol use table functions */

static struct symbol *find_symbol(
        char *nameP) {

        unsigned int index = 0;
        struct symbol *symbolP = symbol_table;

        while (index != first_free_symbol) {
                if (!strcmp(nameP, symbolP->name))
                        return symbolP;

                index++;
                symbolP++;
        }

        return NULL;
}

static struct symbol *check_symbol(
        char *nameP) {

        if (*nameP == '$' || *nameP == '=' || strchr(nameP, ','))
                illegal_character();

        if (strlen(nameP) > MAX_SYMBOL_LENGTH)
                too_long_symbol(nameP);

        return find_symbol(nameP);
}

static struct symbol *allocate_symbol(void) {

        if (first_free_symbol == NUM_SYMBOLS)
                symbol_table_full();

        return symbol_table + first_free_symbol++;
}

static struct symbol_use *allocate_symbol_use(void) {

        if (first_free_symbol_use == NUM_SYMBOL_USES)
                symbol_use_table_full();

        return symbol_use_table + first_free_symbol_use++;
}

static struct symbol *add_symbol(
        char *nameP,
        unsigned int value) {

        struct symbol *symbolP;

        if (!(symbolP = check_symbol(nameP))) {
                symbolP = allocate_symbol();

                strcpy(symbolP->name, nameP);
        }

        if (symbolP->satisfied && symbolP->value != value)
                duplicate_symbol(nameP);

        symbolP->satisfied = 1;
        symbolP->value = value;

        return symbolP;
}

static struct symbol *add_symbol_use(
        char *nameP) {

        struct symbol *symbolP;
        struct symbol_use *symbol_useP;

        if (!(symbolP = check_symbol(nameP))) {
                symbolP = allocate_symbol();

                strcpy(symbolP->name, nameP);
        }

        symbol_useP = allocate_symbol_use();

        symbol_useP->next_useP = symbolP->useP;
        symbolP->useP = symbol_useP;

        return symbolP;
}


/* General utility functions */

static unsigned int charconv(
        unsigned int ch) {

        if (ch < 0x20 || ch > 0x6F)
                illegal_character();

        if (ch <= 0x3F)
                return ch;

        if (ch <= 0x5F)
                return ch - 0x40;

        return ch - 0x20;
}

static unsigned int read_hex(
        char *charP) {

        unsigned int ch;
        unsigned int length = 0;
        unsigned int value = 0;

        while (ch = *(unsigned char *) charP++) {
                length++;
                value <<= 4;

                if ('0' <= ch && ch <= '9')
                        value += ch - '0';
                else if ('A' <= ch && ch <= 'F')
                        value += ch - 'A' + 10;
                else
                        illegal_character();
        }

        if (length > 4)
                too_long_hex_number();

        return value;
}

static unsigned int read_symconstant(void) {

        struct symbol *symbolP;
        char *nameP = paramP;

        if (nameP && *nameP == '{') {
                nameP = strcpy(name, nameP + 1);
                nameP[strlen(nameP) - 1] = 0;

                if (!*nameP)
                        nameP = NULL;
        }

        if (!nameP)
                missing_parameter();

        if (*nameP == '$') {
                if (!*(nameP + 1))
                        return org;

                else if (*(nameP + 1) == '+')
                        return (org + read_hex(nameP + 2)) & 0xFFFF;

                else if (*(nameP + 1) == '-')
                        return (org - read_hex(nameP + 2)) & 0xFFFF;

                else
                        org_rel_syntax_error();
        }

        else if (*nameP == '=')
                return read_hex(nameP + 1);

        if (!second_pass)
                return org;

        symbolP = add_symbol_use(nameP);

        if (!symbolP->satisfied)
                unsatisfied_symbol(nameP);

        return symbolP->value;
}

static char *next_word(
        char *wordP) {

        while (*wordP && *wordP != ' ' && *wordP != '\t' && *wordP != '.')
                wordP++;

        if (!*wordP)
                return NULL;

        if (*wordP == '.')
                return wordP;

        *wordP++ = 0;

        while (*wordP == ' ' || *wordP == '\t')
                wordP++;

        if (!*wordP)
                return NULL;

        return wordP;
}


/* Printing function */

static void print_line(void) {

        if (quiet)
                return;

        if (org_defined)
                sprintf(output, "%04X ", org);
        else
                sprintf(output, "     ");

        if (added) {
                sprintf(output + 5, "%03X +", code);

                if (addedinstrP)
                        sprintf(output + 10, "        %s", addedinstrP);

                if (addedparamP)
                        sprintf(output + strlen(output), " %s", addedparamP);
        }

        else {
                if (directive)
                        sprintf(output + 5, "     %s", instrP);
                else
                        sprintf(output + 5, "%03X          %s", code, instrP);

                if (paramP)
                        sprintf(output + strlen(output), " %s", paramP);

                if (commentP)
                        sprintf(output + strlen(output), "%*s.%s",
                                (int) strlen(output) < COMMENT_COLUMN - 1 ?
                                COMMENT_COLUMN - 1 - (int) strlen(output) : 0, "",
                                commentP);
        }

        if (addedcommentP)
                sprintf(output + strlen(output), "%*s.%s",
                        (int) strlen(output) < COMMENT_COLUMN - 1 ?
                        COMMENT_COLUMN - 1 - (int) strlen(output) : 0, "",
                        addedcommentP);

        if (org_defined && !directive && org >> 12 == pagenum) {
                if (object[(org & 0xFFF) << 1] ||
                    object[((org & 0xFFF) << 1) + 1])
                        overwriting_object_data();

                object[(org & 0xFFF) << 1] = (code | ctrlbits) >> 8;
                object[((org & 0xFFF) << 1) + 1] = code;
        }

        fprintf(outputP, "%s\n", output);
}


/* Directive handling functions */

static void handle_ORG(void) {

        if (!paramP)
                missing_parameter();

        org_defined = 1;
        org = read_hex(paramP);
}

static void handle_LABEL(void) {

        if (!paramP)
                missing_parameter();

        if (!org_defined)
                before_ORG();

        if (second_pass)
                return;

        add_symbol(paramP, org);
}

static void handle_GLABEL(void) {

        struct symbol *symbolP;

        if (!paramP)
                missing_parameter();

        if (!org_defined)
                before_ORG();

        if (second_pass)
                return;

        symbolP = add_symbol(paramP, org);

        symbolP->global = 1;
}

static void handle_EQU(void) {

        char *valueP;

        if (!paramP)
                missing_parameter();

        if (second_pass)
                return;

        if (!(valueP = strchr(paramP, ',')))
                missing_parameter();

        *valueP++ = 0;

        if (!*paramP || !*valueP)
                missing_parameter();

        add_symbol(paramP, read_hex(valueP));
}

static void handle_GEQU(void) {

        char *valueP;
        struct symbol *symbolP;

        if (!paramP)
                missing_parameter();

        if (second_pass)
                return;

        if (!(valueP = strchr(paramP, ',')))
                missing_parameter();

        *valueP++ = 0;

        if (!*paramP || !*valueP)
                missing_parameter();

        symbolP = add_symbol(paramP, read_hex(valueP));

        symbolP->global = 1;
}

static void handle_NAME(void) {

        char *charP;
        unsigned int ch;
        unsigned int length;

        if (!paramP)
                missing_parameter();

        if (!org_defined)
                before_ORG();

        if ((length = strlen(paramP)) > 12)
                too_long_symbol(paramP);

        directive = 0;
        added = 1;

        addedinstrP = addinstr;
        strcpy(addinstr, "CON");

        addedparamP = addparam;

        addedcommentP = addcomment;
        addcomment[1] = 0;

        charP = paramP + length;

        ch = *(unsigned char *) --charP;
        code = charconv(ch) | 0x80;

        sprintf(addparam, "=%03X", code);
        addcomment[0] = ch;

        print_line();

        org = (org + 1) & 0xFFFF;

        while (--length) {
                ch = *(unsigned char *) --charP;
                code = charconv(ch);

                sprintf(addparam, "=%03X", code);
                addcomment[0] = ch;

                print_line();

                org = (org + 1) & 0xFFFF;
        }

        if (second_pass)
                return;

        strncpy(name, paramP, MAX_SYMBOL_LENGTH);
        name[MAX_SYMBOL_LENGTH] = 0;

        add_symbol(name, org);
}

static void handle_MESSAGE(void) {

        char *charP;
        unsigned int ch;
        unsigned int length;

        if (!paramP)
                missing_parameter();

        if (!org_defined)
                before_ORG();

        length = strlen(paramP);

        directive = 0;
        added = 1;

        addedinstrP = addinstr;
        strcpy(addinstr, "CON");

        addedparamP = addparam;

        addedcommentP = addcomment;
        addcomment[1] = 0;

        charP = paramP;

        ch = *(unsigned char *) charP++;
        code = charconv(ch) | 0x100;

        sprintf(addparam, "=%03X", code);
        addcomment[0] = ch;

        print_line();

        org = (org + 1) & 0xFFFF;

        while (--length) {
                ch = *(unsigned char *) charP++;
                code = charconv(ch);

                sprintf(addparam, "=%03X", code);
                addcomment[0] = ch;

                print_line();

                org = (org + 1) & 0xFFFF;
        }
}

static void handle_DMESSAGE(void) {

        char *charP;
        unsigned int ch;
        unsigned int length;

        if (!paramP)
                missing_parameter();

        if (!org_defined)
                before_ORG();

        length = strlen(paramP);

        directive = 0;
        added = 1;

        addedinstrP = addinstr;
        strcpy(addinstr, "CON");

        addedparamP = addparam;

        addedcommentP = addcomment;
        addcomment[1] = 0;

        charP = paramP;

        while (length--) {
                ch = *(unsigned char *) charP++;
                code = charconv(ch);

                if (!length)
                        code |= 0x200;

                sprintf(addparam, "=%03X", code);
                addcomment[0] = ch;

                print_line();

                org = (org + 1) & 0xFFFF;
        }
}

static void handle_INCLUDE(void) {

        if (!paramP)
                missing_parameter();

        handle_file(paramP);
}

static void handle_EXPORT(void) {

        FILE *fileP;
        unsigned int index = 0;
        struct symbol *symbolP = symbol_table;

        if (!paramP)
                missing_parameter();

        if (!second_pass)
                return;

        if (!(fileP = fopen(paramP, "w")))
                file_open_write_error(paramP);

        while (index != first_free_symbol) {
                if (symbolP->global)
                        fprintf(fileP, "%%EQU {%s,%04X}\n",
                                symbolP->name, symbolP->value);

                index++;
                symbolP++;
        }

        fclose(fileP);
}

static void handle_QUIET_ON(void) {

        if (paramP)
                extra_parameter();

        quiet = 1;
}

static void handle_QUIET_OFF(void) {

        if (paramP)
                extra_parameter();

        quiet = 0;
}

static void handle_INITPAGE(void) {

        if (!paramP)
                missing_parameter();

        if (strlen(paramP) > 1)
                too_long_page_number();

        pagenum = read_hex(paramP);

        if (!second_pass)
                return;

        memset(object, 0, sizeof object);
}

static void handle_OBJDUMP(void) {

        FILE *fileP;
        unsigned int index;
        int cks = 0x3FF;
        unsigned int crc = -1;
        unsigned int data;
        unsigned int bit;
        unsigned int i;

        if (!paramP)
                missing_parameter();

        if (pagenum < 0)
                object_page_not_defined();

        org_defined = 1;
        org = pagenum << 12 | 0xFFF;

        if (!second_pass)
                return;

        /* Count ten-bit check sum */
        index = 0;

        while (index < sizeof object - 2) {
                data = object[index++] & 0x3;
                data <<= 8;
                data |= object[index++];

                cks -= data;
                if (cks < 0)
                        cks += 0x3FF;
        }

        cks = (cks + 1) % 0x400;
        code = cks;
        ctrlbits = 0;

        directive = 0;
        added = 1;

        addedinstrP = addinstr;
        strcpy(addinstr, "CON");

        addedparamP = addparam;
        sprintf(addparam, "=%03X", code);

        addedcommentP = addcomment;
        strcpy(addcomment, "Check sum");

        print_line();

        /* Count 32-bit CRC */
        index = 0;

        while (index < sizeof object) {
                data = object[index++];
                data <<= 8;
                data |= object[index++];

                i = 0;

                while (i++ < 16) {
                        data &= (1 << 16) - 1;
                        data <<= 1;

                        bit = crc >> 31;
                        crc <<= 1;

                        if ((data >> 16) != bit)
                                crc ^= 0x04C11DB7;
                }
        }

        fprintf(outputP, ".CRC = %08X\n", crc);

        if (!(fileP = fopen(paramP, "wb")))
                file_open_write_error(paramP);

        index = 0;

        while (index < sizeof object)
                fputc(object[index++], fileP);

        fclose(fileP);
}

static void handle_TURBODIS(void) {

        if (paramP)
                extra_parameter();

        ctrlbits = 0x3000;
}

static void handle_TURBOEN(void) {

        if (paramP)
                extra_parameter();

        ctrlbits = 0;
}

static void handle_directive(void) {

        if (!*instrP)
                empty_directive();

        if (paramP && *paramP == '{') {
                paramP++;
                paramP[strlen(paramP) - 1] = 0;

                if (!*paramP)
                        paramP = NULL;
        }

        if (!strcmp(instrP, "ORG"))
                handle_ORG();
        else if (!strcmp(instrP, "LABEL"))
                handle_LABEL();
        else if (!strcmp(instrP, "GLABEL"))
                handle_GLABEL();
        else if (!strcmp(instrP, "EQU"))
                handle_EQU();
        else if (!strcmp(instrP, "GEQU"))
                handle_GEQU();
        else if (!strcmp(instrP, "NAME"))
                handle_NAME();
        else if (!strcmp(instrP, "MESSAGE"))
                handle_MESSAGE();
        else if (!strcmp(instrP, "DMESSAGE"))
                handle_DMESSAGE();
        else if (!strcmp(instrP, "INCLUDE"))
                handle_INCLUDE();
        else if (!strcmp(instrP, "EXPORT"))
                handle_EXPORT();
        else if (!strcmp(instrP, "QUIET_ON"))
                handle_QUIET_ON();
        else if (!strcmp(instrP, "QUIET_OFF"))
                handle_QUIET_OFF();
        else if (!strcmp(instrP, "INITPAGE"))
                handle_INITPAGE();
        else if (!strcmp(instrP, "OBJDUMP"))
                handle_OBJDUMP();
        else if (!strcmp(instrP, "TURBODIS"))
                handle_TURBODIS();
        else if (!strcmp(instrP, "TURBOEN"))
                handle_TURBOEN();
        else
                unknown_directive();
}


/* Code generation functions */

static void assemble(
        const struct descriptor *descriptorP) {

        const char **params;
        unsigned int index;
        int distance;
        unsigned int target;

        switch (descriptorP->param_type) {
        case PARAM_TYPE_NONE:
                if (paramP)
                        extra_parameter();

                code = descriptorP->instr << 2 | descriptorP->param << 6;

                return;

        case PARAM_TYPE_IMM_WORD:
                added = 1;

                code = descriptorP->instr << 2 | descriptorP->param << 6;

                print_line();

                org = (org + 1) & 0xFFFF;

                added = 0;

        case PARAM_TYPE_CONST:
                code = read_symconstant() & 0x3FF;

                return;

        case PARAM_TYPE_D:
                if (!paramP)
                        missing_parameter();

                params = D_param;
                index = 0;

                while (index < 16) {
                        if (!strcmp(paramP, *params)) {
                                code = descriptorP->instr << 2 | index << 6;

                                return;
                        }

                        params++;
                        index++;
                }

                break;

        case PARAM_TYPE_N:
                if (!paramP)
                        missing_parameter();

                params = N_param;
                index = 0;

                while (index < 16) {
                        if (!strcmp(paramP, *params)) {
                                code = descriptorP->instr << 2 | index << 6;

                                return;
                        }

                        params++;
                        index++;
                }

                break;

        case PARAM_TYPE_TEF:
                if (!paramP)
                        missing_parameter();

                params = TEF_param;
                index = 0;

                while (index < 8) {
                        if (!strcmp(paramP, *params)) {
                                code = descriptorP->instr << 5 | index << 2 | 2;

                                return;
                        }

                        params++;
                        index++;
                }

                break;

        case PARAM_TYPE_REL:
                distance = read_symconstant() - org;

                if (distance < -0x40 || distance > 0x3F)
                        too_long_jump();

                if (*paramP != '$') {
                        addedcommentP = addcomment;
                        if (distance >= 0)
                                sprintf(addcomment, "$+%02X", distance);
                        else
                                sprintf(addcomment, "$-%02X", -distance);
                }

                code = descriptorP->instr << 2 |
                        ((unsigned int) distance & 0x7F) << 3 | 3;

                return;

        case PARAM_TYPE_ABS:
                target = read_symconstant();

                added = 1;

                code = (target & 0xFF) << 2 | 1;

                print_line();

                org = (org + 1) & 0xFFFF;

                if (*paramP != '=') {
                        addedcommentP = addcomment;
                        sprintf(addcomment, "=%04X", target);
                }

                added = 0;

                code = (target & 0xFF00) >> 6 | descriptorP->instr;

                return;

        case PARAM_TYPE_DCONST:
                target = (read_symconstant() - (org & 0xF000)) & 0xFFFF;

                added = 1;

                code = (target >> 8) | (descriptorP->instr << 8);

                print_line();

                org = (org + 1) & 0xFFFF;

                added = 0;

                code = target & 0xFF;

                return;
        }

        parameter_not_found();
}

static void find_descriptor(void) {

        const struct descriptor *descriptorP;
        unsigned int index;

        if (!org_defined)
                before_ORG();

        descriptorP = CON_descriptor;
        index = 0;

        while (index < sizeof CON_descriptor / sizeof (struct descriptor)) {
                if (!strcmp(instrP, descriptorP->mnem)) {
                        assemble(descriptorP);

                        return;
                }

                descriptorP++;
                index++;
        }

        descriptorP = class0_descriptor;
        index = 0;

        while (index < sizeof class0_descriptor / sizeof (struct descriptor)) {
                if (!strcmp(instrP, descriptorP->mnem)) {
                        assemble(descriptorP);

                        return;
                }

                descriptorP++;
                index++;
        }

        descriptorP = class1_descriptor;
        index = 0;

        while (index < sizeof class1_descriptor / sizeof (struct descriptor)) {
                if (!strcmp(instrP, descriptorP->mnem)) {
                        assemble(descriptorP);

                        return;
                }

                descriptorP++;
                index++;
        }

        descriptorP = class2_descriptor;
        index = 0;

        while (index < sizeof class2_descriptor / sizeof (struct descriptor)) {
                if (!strcmp(instrP, descriptorP->mnem)) {
                        assemble(descriptorP);

                        return;
                }

                descriptorP++;
                index++;
        }

        descriptorP = class3_descriptor;
        index = 0;

        while (index < sizeof class3_descriptor / sizeof (struct descriptor)) {
                if (!strcmp(instrP, descriptorP->mnem)) {
                        assemble(descriptorP);

                        return;
                }

                descriptorP++;
                index++;
        }

        instruction_not_found();
}


/* Line handling function */

static void handle_line(void) {

        char *eolP;
        char *extraP;

        if (eolP = strchr(line, '\n'))
                *eolP = 0;
        else
                too_long_line();

        added = 0;
        instrP = line;
        commentP = NULL;
        addedinstrP = NULL;
        addedparamP = NULL;
        addedcommentP = NULL;

        while (*instrP == ' ' || *instrP == '\t')
                instrP++;

        if (!*instrP || *instrP == '.') {
                if (!quiet)
                        fprintf(outputP, "%s\n", instrP);

                return;
        }

        if (paramP = next_word(instrP)) {
                if (*paramP == '.') {
                        commentP = paramP;
                        *commentP++ = 0;
                        paramP = NULL;
                }

                else if (*paramP == '{') {
                        if (!(extraP = strchr(paramP + 1, '}')))
                                brace_error();

                        if (commentP = strchr(++extraP, '.'))
                                *commentP++ = 0;

                        if (*extraP && *extraP != ' ' && *extraP != '\t' ||
                            strchr(paramP + 1, '{'))
                                brace_error();

                        if (extraP = next_word(extraP))
                                extra_second_parameter();
                }

                else {
                        if (commentP = strchr(paramP, '.'))
                                *commentP++ = 0;

                        if (strchr(paramP + 1, '{') || strchr(paramP, '}'))
                                brace_error();

                        if (extraP = next_word(paramP))
                                extra_second_parameter();
                }
        }

        if (*instrP == '%') {
                directive = 1;

                print_line();

                instrP++;
                handle_directive();

                return;
        }

        directive = 0;

        find_descriptor();

        print_line();

        if (org_defined)
                org = (org + 1) & 0xFFFF;
}


/* File handling function */

static void handle_file(
        char *filenameP) {

        FILE *fileP;

        if (!(fileP = fopen(filenameP, "r")))
                file_open_read_error(filenameP);

        while (fgets(line, MAX_LINE_LEN + 1, fileP))
                handle_line();

        fclose(fileP);
}


/* Main function */

int main(
        int argc,
        char **argv) {

        if (argc < 2) {
                fprintf(stderr, "Usage: %s asmfile\n", argv[0]);
                return 1;
        }

        outputP = stderr;
        pagenum = -1;

        handle_file(argv[1]);

        second_pass = 1;
        quiet = 0;
        org_defined = 0;
        outputP = stdout;
        pagenum = -1;
        ctrlbits = 0;

        handle_file(argv[1]);

        return 0;
}
